require_relative "version"
require_relative "attributes/ext/boolean"

module Toolchain::Attributes
  require_relative "attributes/configuration"
  require_relative "attributes/errors"
  require_relative "attributes/helpers"

  module ClassMethods

    # @return [Array<Symbol>] all defined attributes.
    #
    def keys
      @keys ||= []
    end

    # Defines an attribute on the Class.
    #
    # @example
    #   class Company
    #     attribute :name, String, "Unnamed"
    #     attribute :email, String
    #   end
    #
    # @param key [Symbol, String]
    # @param type [Class]
    # @param default [Object]
    #
    def attribute(key, type, default = nil)
      type = [TrueClass, FalseClass] if type == Boolean

      if Helpers.invalid_value?(default, type)
        raise Errors::TypeMismatch,
          "expected #{self.name}##{key} to have default value " +
          "of #{type} type, but received #{default.class} (#{default})."
      end

      keys.push(key.to_sym)

      define_method(key) do
        value = instance_variable_get("@#{key}")

        if value.nil?
          new_value =
            begin
              if default.kind_of?(Proc)
                default.call
              else
                default.dup
              end
            rescue TypeError
              default
            end

          send("#{key}=", new_value)
        else
          value
        end
      end

      define_method("#{key}=") do |value|
        if value.kind_of?(String) && type.respond_to?(:parse)
          value = type.parse(value)
        end

        if type == [TrueClass, FalseClass] && [0, "0"].include?(value)
          value = false
        end

        if type == [TrueClass, FalseClass] && [1, "1"].include?(value)
          value = true
        end

        if Helpers.invalid_value?(value, type)
          raise Errors::TypeMismatch,
            "#{self.class}##{key} expected #{type} type, " +
            "but received #{value.class} (#{value})."
        end

        if value.kind_of?(Hash)
          transformation = Configuration.hash_transformation
          value = Helpers.send(transformation, value)
        end

        instance_variable_set("@#{key}", value)
      end
    end

    # Takes a Proc that contains attribute definitions and applies
    # that to this class.
    #
    # @param attributes [Proc]
    #
    def include_attributes(attributes)
      class_eval(&attributes)
    end
  end

  module InstanceMethods

    # @param attributes [Hash]
    #
    def initialize(attributes = {})
      self.attributes = attributes
    end

    # @return [Hash] keys and values for each defined attribute.
    #
    def attributes
      include_nil = Configuration.include_nil_in_attributes

      attributes = Hash.new.tap do |attrs|
        Helpers.each_key(self.class) do |key|
          value = send(key)

          if !value.nil? || (value.nil? && include_nil)
            attrs[key] = value
          end
        end
      end

      transformation = Configuration.hash_transformation
      Helpers.send(transformation, attributes)
    end

    # Mass-assignment for the defined attributes by passing
    # in a Hash. Non-existing attributes are ignored.
    #
    # @param value [Hash]
    #
    def attributes=(value)
      if !value.kind_of?(Hash)
        raise Errors::InvalidMassAssignment,
          "Can't mass-assign #{value.class} (#{value}) type " +
          "to #{self.class}#attributes."
      end

      value = Helpers.symbolize_keys(value)
      keys = Array.new.tap do |keys|
        Helpers.each_key(self.class) { |key| keys << key }
        keys.uniq!
      end

      value.each do |key, value|
        send("#{key}=", value) if keys.include?(key)
      end
    end
  end

  def self.included(base)
    base.send(:include, InstanceMethods)
    base.extend(ClassMethods)
  end
end
