require_relative "version"

module Toolchain::Validations
  require_relative "validations/helpers"
  require_relative "validations/validation_errors"
  require_relative "validations/validators"
  require_relative "validations/delegator"
  require_relative "validations/delegation"

  module ClassMethods

    # @return [Array] all the standard validation operations to run.
    #
    def validations
      @validations ||= []
    end

    # @return [Array] all the custom validation operations to run.
    #
    def custom_validations
      @custom_validations ||= []
    end

    # Adds the provided validator to the validations array for
    # the provided attribute key_path.
    #
    # @param args [Array<Symbol, Hash>]
    #
    # @example
    #   class Device
    #     validates :name, presence: true
    #     validates :config, :url, presence: true
    #   end
    #
    def validates(*args)
      options, key_path = args.pop, args

      _if = options.delete(:if)
      _unless = options.delete(:unless)

      options.each do |key, data|
        if validator = Helpers.find_validator(key)
          validations.push(
            key_path: key_path,
            validator: validator,
            data: data,
            if: _if,
            unless: _unless
          )
        end
      end
    end

    # Adds the provided method name to the validations array.
    # This method will be called during validation time and allows
    # you to do program-specific validations.
    #
    # @param method [Symbol, String]
    #
    def validate(method)
      custom_validations.push(method.to_sym)
    end

    # Takes a Proc that contains validation logic and applies
    # that to this class.
    #
    # @param validations [Proc]
    #
    def include_validations(validations)
      class_eval(&validations)
    end
  end

  module InstanceMethods

    # @return [Toolchain::Validations::ValidationErrors]
    #
    def errors
      @errors ||= ValidationErrors.new
    end

    # Runs all the validation operations and will apply error
    # messages to the ValidationErrors object when validator errors occur.
    #
    def validate
      Helpers.each_validation(self.class) do |v|
        validate = true

        if v[:if].kind_of?(Proc) && v[:if].call(self) == false
          validate = false
        end

        if v[:unless].kind_of?(Proc) && v[:unless].call(self) == true
          validate = false
        end

        if validate
          v[:validator].new(self, v[:key_path], v[:data]).validate
        end
      end

      Helpers.each_validation(self.class, :custom_validations) { |v| send(v) }
    end

    # Resets the ValidationErrors object, re-validates the current object
    # and then determines whether or not this object is valid.
    #
    # @note If invalid (returns `false`), errors will have been
    #   added to the `object.errors` class.
    #
    # @return [Boolean]
    #
    def valid?
      errors.reset
      validate
      errors.empty?
    end
  end

  def self.included(base)
    base.send(:include, InstanceMethods)
    base.extend(ClassMethods)
  end
end
