module Toolchain::Validations::Helpers
  extend self

  # Finds the validator by key name.
  #
  # @param key [Symbol, String]
  #
  # @return [Object, nil] nil if the validator doesn't exist.
  #
  def find_validator(key)
    Toolchain::Validations::Validators.const_get(classify(key))
  rescue NameError
  end

  # Yields all defined validations from the current class,
  # as well as all the (relavant) super-classes.
  #
  # @param klass [Class]
  # @param method [Symbol, String]
  #
  # @yield [Hash]
  #
  def each_validation(klass, method = :validations)
    while ![Class, Module, Object, BasicObject, nil].include?(klass)
      klass.send(method).each { |v| yield v } if klass.respond_to?(method)
      klass = klass.superclass
    end
  end

  # Converts the provided value to it's class name.
  #
  # @param value [Symbol, String]
  #
  # @return [String]
  #
  def classify(value)
    value.to_s.split("_").map(&:capitalize).join
  end

  # Custom inject method that includes a 3rd yield argument
  # that determines whether the current iteration is the last.
  #
  # @param array [Array]
  # @param init [Object] uses first value of array if nil.
  #
  # @yield memo [Object], value [Object], last [Boolean]
  #
  # @return [Object] The final memo result.
  #
  def inject(array, init = nil)
    memo = init || array.shift
    length = array.length
    array.each_with_index do |value, index|
      memo = yield memo, value, index == length - 1
    end
    memo
  end
end
